@@ -28,6 +28,7 @@
  *	More (major) cleanups and bug fixes by Salvador Abreu <spa@fct.unl.pt>
  *	Additional options Lutz Pre"sler <Lutz.Pressler@med-stat.gwdg.de>
  *	Masquerade client support added <Alan.Cox@linux.org>
+ *	Made to match 1.1.91, and Urgen's newer ipfw kernel code.
  */
 
 #include <sys/types.h>
@@ -38,6 +39,8 @@
 #endif
 #endif
 #include <sys/socket.h>
+#include <fcntl.h>
+#include <unistd.h>
 #define IPFIREWALL
 #include <netinet/in.h>
 #include <netinet/ip.h>
@@ -46,7 +49,7 @@
 #ifndef LINUX
 #include <netinet/ip_fw.h>
 #else
-#include <linux/ip_fw.h>	/* Until it gets into stdinc */
+#include <linux/ip_fw.h>	/* Until it gets into stdinc (I've submitted a netinet/ip_fw.h) */
 #endif
 #include <netdb.h>
 #include <stdio.h>
@@ -77,20 +80,25 @@ static int lookup = 1;
 
 void show_usage()
 {
-	fprintf(stderr, "usage: ipfirewall [-n] l[ist]\n");
+	fprintf(stderr, "usage: ipfirewall [-n]\n\t\t  l[ist]  a[ccounting]\n");
+	fprintf(stderr, "\t\t| l[ist]  b[locking]\n");
+	fprintf(stderr, "\t\t| l[ist]  f[irewall]\n");
 	fprintf(stderr, "\t\t| f[lush] a[ccounting]\n");
+	fprintf(stderr, "\t\t| f[lush] b[locking]\n");
 	fprintf(stderr, "\t\t| f[lush] f[irewall]\n");
-	fprintf(stderr, "\t\t| c[heck] b[locking]   <type> from <src> to <dst>\n");
+	fprintf(stderr, "\t\t| c[heck] b[locking] <type> from <src> to <dst>\n");
 	fprintf(stderr, "\t\t| c[heck] f[orwarding] <type> from <src> to <dst>\n");
-	fprintf(stderr, "\t\t| a[dd]   a[ccounting] <type> from <src> to <dst>\n");
-	fprintf(stderr, "\t\t| a[dd]   b[locking]   <type> from <src> to <dst>\n");
-	fprintf(stderr, "\t\t| a[dd]   f[orwarding] <type> from <src> to <dst>\n");
+	fprintf(stderr, "\t\t| p[olicy] b[locking] <accept|deny|reject>\n");
+	fprintf(stderr, "\t\t| p[olicy] f[orwarding] <accept|deny|reject>\n");
+	fprintf(stderr, "\t\t| a[dd]   a[ccounting] <type> [iface <addr>] from <src> to <dst>\n");
+	fprintf(stderr, "\t\t| a[dd]   b[locking]   <type> [iface <addr>] from <src> to <dst>\n");
+	fprintf(stderr, "\t\t| a[dd]   f[orwarding] <type> [iface <addr>] from <src> to <dst>\n");
 #ifdef DO_MASQUERADE	
 	fprintf(stderr, "\t\t| a[dd]   m[asquerade] <type> from <src> to <dst>\n");
 #endif	
-	fprintf(stderr, "\t\t| d[el]   a[ccounting] <type> from <src> to <dst>\n");
-	fprintf(stderr, "\t\t| d[el]   b[locking]   <type> from <src> to <dst>\n");
-	fprintf(stderr, "\t\t| d[el]   f[orwarding] <type> from <src> to <dst>\n");
+	fprintf(stderr, "\t\t| d[el]   a[ccounting] <type> [iface <addr>] from <src> to <dst>\n");
+	fprintf(stderr, "\t\t| d[el]   b[locking]   <type> [iface <addr>] from <src> to <dst>\n");
+	fprintf(stderr, "\t\t| d[el]   f[orwarding] <type> [iface <addr>] from <src> to <dst>\n");
 #ifdef DO_MASQUERADE	
 	fprintf(stderr, "\t\t| d[el]   m[asquerade] <type> from <src> to <dst>\n");
 #endif
@@ -105,7 +113,7 @@ void show_help()
 {
 	show_usage();
 	fprintf(stderr, "where:\n");
-	fprintf(stderr, "       <src> ::= <host>:<port> /* for TCP or UDP */\n");
+	fprintf(stderr, "       <src> ::= <host> <port> /* for TCP or UDP */\n");
 	fprintf(stderr, "       <src> ::= <host>        /* for ICMP */\n");
 	fprintf(stderr, "      <host> ::= <byte>.<byte>.<byte>.<byte>[/<width>]\n");
 	fprintf(stderr, "               | <hostname>\n");
@@ -168,13 +176,16 @@ static int do_setsockopt(char *what, int fd, int proto, int cmd, void *data, int
 
 	switch (cmd)
 	{
-		CASE(FW_FLUSH);
+		CASE(FW_FLUSH_BLK);
+		CASE(FW_FLUSH_FWD);
 		CASE(FW_CHK_BLK);
 		CASE(FW_CHK_FWD);
 		CASE(FW_ADD_BLK);
 		CASE(FW_ADD_FWD);
 		CASE(FW_DEL_BLK);
 		CASE(FW_DEL_FWD);
+		CASE(FW_POLICY_FWD);
+		CASE(FW_POLICY_BLK);
 		CASE(ACCT_ADD);
 		CASE(ACCT_DEL);
 		CASE(ACCT_FLUSH);
@@ -217,7 +228,7 @@ static int do_setsockopt(char *what, int fd, int proto, int cmd, void *data, int
 			struct ip_fw *fp = (struct ip_fw *) data;
 			int fmt_ports = 0;
 			printf("data = struct ip_fw {\n");
-			if (fp->flags & IP_FW_F_ACCEPT)
+			if (fp->fw_flg & IP_FW_F_ACCEPT)
 			{
 				printf("\taccept ");
 			}
@@ -225,7 +236,7 @@ static int do_setsockopt(char *what, int fd, int proto, int cmd, void *data, int
 			{
 				printf("\tdeny ");
 			}
-			switch (fp->flags & IP_FW_F_KIND)
+			switch (fp->fw_flg & IP_FW_F_KIND)
 			{
 			case IP_FW_F_ALL:
 				printf("\tuniversal\n");
@@ -244,28 +255,28 @@ static int do_setsockopt(char *what, int fd, int proto, int cmd, void *data, int
 				fmt_ports = 0;
 				break;
 			}
-			printf("\tsrc=%s:", fmtip(fp->src.s_addr));
-			printf("%s ", fmtip(fp->src_mask.s_addr));
+			printf("\tsrc=%s:", fmtip(fp->fw_src.s_addr));
+			printf("%s ", fmtip(fp->fw_smsk.s_addr));
 			if (fmt_ports)
 			{
-				print_ports(fp->n_src_p, fp->flags & IP_FW_F_SRNG, &fp->ports[0]);
+				print_ports(fp->fw_nsp, fp->fw_flg & IP_FW_F_SRNG, &fp->fw_pts[0]);
 			}
-			else if (fp->flags & (IP_FW_F_SRNG | IP_FW_F_DRNG))
+			else if (fp->fw_flg & (IP_FW_F_SRNG | IP_FW_F_DRNG))
 			{
 				exit(1);
 			}
-			else if (fp->n_src_p > 0 || fp->n_dst_p > 0)
+			else if (fp->fw_nsp > 0 || fp->fw_ndp > 0)
 			{
 				exit(1);
 			}
 			printf("\n");
-			printf("\tdst=%s:", fmtip(fp->dst.s_addr));
-			printf("%s ", fmtip(fp->dst_mask.s_addr));
+			printf("\tdst=%s:", fmtip(fp->fw_dst.s_addr));
+			printf("%s ", fmtip(fp->fw_dmsk.s_addr));
 			if (fmt_ports)
 			{
-				print_ports(fp->n_dst_p,
-				    fp->flags & IP_FW_F_DRNG,
-				    &fp->ports[fp->n_src_p]);
+				print_ports(fp->fw_ndp,
+				    fp->fw_flg & IP_FW_F_DRNG,
+				    &fp->fw_pts[fp->fw_nsp]);
 			}
 			printf("\n");
 			printf("}\n");
@@ -762,10 +773,12 @@ void add(ipf_kind kind, int socket_fd, char **argv)
 		{
 			accept_firewall = IP_FW_F_ACCEPT;
 		}
+		else if (strcmp(*argv, "reject") == 0)
+			accept_firewall = IP_FW_F_ICMPRPL;
 		else
 		{
 			fprintf(stderr,
-			    "ipfw: expected \"accept\" or \"deny\", got \"%s\"\n",
+			    "ipfw: expected \"accept\", \"deny\" or \"reject\", got \"%s\"\n",
 			    *argv);
 			exit(1);
 		}
@@ -773,10 +786,34 @@ void add(ipf_kind kind, int socket_fd, char **argv)
 		argv += 1;
 	}
 	else
-		accept_firewall = 1;
+		accept_firewall = IP_FW_F_ACCEPT;
+	if(*argv==NULL)
+	{
+		fprintf(stderr,"ipfw: missing protocol name.\n");
+		exit(1);
+	}
 	proto_name = *argv++;
 	protocol = get_protocol(proto_name, add_usage, kind);
-
+	
+	firewall.fw_via.s_addr = 0;
+	
+	if(*argv && strcmp(*argv,"iface")==0)
+	{
+		argv++;
+		if(*argv==NULL)
+		{
+			fprintf(stderr,"ipfw: missing interface address.\n");
+			exit(1);
+		}
+		firewall.fw_via.s_addr=inet_addr(*argv);
+		if(firewall.fw_via.s_addr==-1)
+		{
+			fprintf(stderr,"Invalid interface address.\n");
+			exit(1);
+		}
+		argv++;
+	}
+	
 	if (*argv == NULL)
 	{
 		fprintf(stderr, "ipfw: missing \"from\" keyword\n");
@@ -785,11 +822,11 @@ void add(ipf_kind kind, int socket_fd, char **argv)
 	if (strcmp(*argv, "from") == 0)
 	{
 		argv++;
-		get_ipaddr(*argv++, &firewall.src, &firewall.src_mask, add_usage, kind);
+		get_ipaddr(*argv++, &firewall.fw_src, &firewall.fw_smsk, add_usage, kind);
 		if (protocol == IP_FW_F_TCP || protocol == IP_FW_F_UDP)
 		{
 			int cnt;
-			cnt = get_ports(&argv, &firewall.ports[0], 0, IP_FW_MAX_PORTS,
+			cnt = get_ports(&argv, &firewall.fw_pts[0], 0, IP_FW_MAX_PORTS,
 			    add_usage, kind, proto_name);
 			if (cnt < 0)
 			{
@@ -800,11 +837,11 @@ void add(ipf_kind kind, int socket_fd, char **argv)
 			{
 				src_range = 0;
 			}
-			firewall.n_src_p = cnt;
+			firewall.fw_nsp = cnt;
 		}
 		else
 		{
-			firewall.n_src_p = 0;
+			firewall.fw_nsp = 0;
 			src_range = 0;
 		}
 	}
@@ -822,12 +859,12 @@ void add(ipf_kind kind, int socket_fd, char **argv)
 	if (strcmp(*argv, "to") == 0)
 	{
 		argv++;
-		get_ipaddr(*argv++, &firewall.dst, &firewall.dst_mask, add_usage, kind);
+		get_ipaddr(*argv++, &firewall.fw_dst, &firewall.fw_dmsk, add_usage, kind);
 		if (protocol == IP_FW_F_TCP || protocol == IP_FW_F_UDP)
 		{
 			int cnt;
-			cnt = get_ports(&argv, &firewall.ports[firewall.n_src_p], 0,
-			    IP_FW_MAX_PORTS - firewall.n_src_p,
+			cnt = get_ports(&argv, &firewall.fw_pts[firewall.fw_nsp], 0,
+			    IP_FW_MAX_PORTS - firewall.fw_nsp,
 			    add_usage, kind, proto_name);
 			if (cnt < 0)
 			{
@@ -838,11 +875,11 @@ void add(ipf_kind kind, int socket_fd, char **argv)
 			{
 				dst_range = 0;
 			}
-			firewall.n_dst_p = cnt;
+			firewall.fw_ndp = cnt;
 		}
 		else
 		{
-			firewall.n_dst_p = 0;
+			firewall.fw_ndp = 0;
 			dst_range = 0;
 		}
 	}
@@ -854,10 +891,10 @@ void add(ipf_kind kind, int socket_fd, char **argv)
 
 	if (*argv == NULL)
 	{
-		firewall.flags = protocol | accept_firewall | src_range | dst_range;
+		firewall.fw_flg = protocol | accept_firewall | src_range | dst_range;
 #ifdef DO_MASQUERADE		
 		if (kind == IPF_MASQUERADE)
-			firewall.flags |= IP_FW_F_MASQ;
+			firewall.fw_flg |= IP_FW_F_MASQ;
 #endif
 		(void) do_setsockopt(ipf_names[kind],
 		    socket_fd, IPPROTO_IP,
@@ -905,6 +942,8 @@ void del(ipf_kind kind, int socket_fd, char **argv)
 		{
 			accept_firewall = IP_FW_F_ACCEPT;
 		}
+		else if (strcmp(*argv, "reject") == 0)
+			accept_firewall = IP_FW_F_ICMPRPL;
 		else
 		{
 			fprintf(stderr,
@@ -916,10 +955,35 @@ void del(ipf_kind kind, int socket_fd, char **argv)
 		argv += 1;
 	}
 	else
-		accept_firewall = 1;
+		accept_firewall = IP_FW_F_ACCEPT;
+	if(*argv==NULL)
+	{
+		fprintf(stderr,"ipfw: missing protocol name.\n");
+		exit(1);
+	}
+		
 	proto_name = *argv++;
 	protocol = get_protocol(proto_name, del_usage, kind);
 
+	firewall.fw_via.s_addr = 0;
+	
+	if(*argv && strcmp(*argv,"iface")==0)
+	{
+		argv++;
+		if(*argv==NULL)
+		{
+			fprintf(stderr,"ipfw: missing interface address.\n");
+			exit(1);
+		}
+		firewall.fw_via.s_addr=inet_addr(*argv);
+		if(firewall.fw_via.s_addr==-1)
+		{
+			fprintf(stderr,"Invalid interface address.\n");
+			exit(1);
+		}
+		argv++;
+	}
+
 	if (*argv == NULL)
 	{
 		fprintf(stderr, "ipfw: missing \"from\" keyword\n");
@@ -928,11 +992,11 @@ void del(ipf_kind kind, int socket_fd, char **argv)
 	if (strcmp(*argv, "from") == 0)
 	{
 		argv++;
-		get_ipaddr(*argv++, &firewall.src, &firewall.src_mask, del_usage, kind);
+		get_ipaddr(*argv++, &firewall.fw_src, &firewall.fw_smsk, del_usage, kind);
 		if (protocol == IP_FW_F_TCP || protocol == IP_FW_F_UDP)
 		{
 			int cnt;
-			cnt = get_ports(&argv, &firewall.ports[0], 0, IP_FW_MAX_PORTS,
+			cnt = get_ports(&argv, &firewall.fw_pts[0], 0, IP_FW_MAX_PORTS,
 			    del_usage, kind, proto_name);
 			if (cnt < 0)
 			{
@@ -943,11 +1007,11 @@ void del(ipf_kind kind, int socket_fd, char **argv)
 			{
 				src_range = 0;
 			}
-			firewall.n_src_p = cnt;
+			firewall.fw_nsp = cnt;
 		}
 		else
 		{
-			firewall.n_src_p = 0;
+			firewall.fw_nsp = 0;
 			src_range = 0;
 		}
 	}
@@ -965,12 +1029,12 @@ void del(ipf_kind kind, int socket_fd, char **argv)
 	if (strcmp(*argv, "to") == 0)
 	{
 		argv++;
-		get_ipaddr(*argv++, &firewall.dst, &firewall.dst_mask, del_usage, kind);
+		get_ipaddr(*argv++, &firewall.fw_dst, &firewall.fw_dmsk, del_usage, kind);
 		if (protocol == IP_FW_F_TCP || protocol == IP_FW_F_UDP)
 		{
 			int cnt;
-			cnt = get_ports(&argv, &firewall.ports[firewall.n_src_p], 0,
-			    IP_FW_MAX_PORTS - firewall.n_src_p,
+			cnt = get_ports(&argv, &firewall.fw_pts[firewall.fw_nsp], 0,
+			    IP_FW_MAX_PORTS - firewall.fw_nsp,
 			    del_usage, kind, proto_name);
 			if (cnt < 0)
 			{
@@ -981,11 +1045,11 @@ void del(ipf_kind kind, int socket_fd, char **argv)
 			{
 				dst_range = 0;
 			}
-			firewall.n_dst_p = cnt;
+			firewall.fw_ndp = cnt;
 		}
 		else
 		{
-			firewall.n_dst_p = 0;
+			firewall.fw_ndp = 0;
 			dst_range = 0;
 		}
 	}
@@ -997,10 +1061,10 @@ void del(ipf_kind kind, int socket_fd, char **argv)
 
 	if (*argv == NULL)
 	{
-		firewall.flags = protocol | accept_firewall | src_range | dst_range;
+		firewall.fw_flg = protocol | accept_firewall | src_range | dst_range;
 #ifdef DO_MASQUERADE
 		if (kind == IPF_MASQUERADE)
-			firewall.flags |= IP_FW_F_MASQ;
+			firewall.fw_flg |= IP_FW_F_MASQ;
 #endif			
 		(void) do_setsockopt(ipf_names[kind],
 		    socket_fd, IPPROTO_IP,
@@ -1029,11 +1093,11 @@ static int count_mask(unsigned long mask)
 typedef struct
 {
 	int acct;
-	unsigned long sa, da, sm, dm;
+	unsigned long sa, da, sm, dm, iface;
 	unsigned int nsp, ndp;
 	unsigned long npkt, nbyt;
-	unsigned int ports[10];
-	int flags;
+	unsigned int fw_pts[10];
+	int fw_flg;
 } fw_rec;
 
 #define MIN(a,b) ((a)<(b)? (a): (b))
@@ -1051,10 +1115,10 @@ static int list_order(const void *L, const void *R)
 	register const fw_rec *r = R;
 
 	register int result =
-	((r->flags & IP_FW_F_KIND) - (l->flags & IP_FW_F_KIND)) ? :
+	((r->fw_flg & IP_FW_F_KIND) - (l->fw_flg & IP_FW_F_KIND)) ? :
 	(MIN(SRC(r), DST(r)) - MIN(SRC(l), DST(l)));
 
-	if (result == 0 && (l->flags & (IP_FW_F_TCP | IP_FW_F_UDP)))
+	if (result == 0 && (l->fw_flg & (IP_FW_F_TCP | IP_FW_F_UDP)))
 	{
 		unsigned int nlp, lp[10], nrp, rp[10];
 		unsigned int i;
@@ -1062,8 +1126,8 @@ static int list_order(const void *L, const void *R)
 		bzero(lp, 10 * sizeof(unsigned int));
 		bzero(rp, 10 * sizeof(unsigned int));
 
-		bcopy(l->ports, lp, (nlp = l->nsp + l->ndp) * sizeof(unsigned int));
-		bcopy(r->ports, rp, (nrp = r->nsp + r->ndp) * sizeof(unsigned int));
+		bcopy(l->fw_pts, lp, (nlp = l->nsp + l->ndp) * sizeof(unsigned int));
+		bcopy(r->fw_pts, rp, (nrp = r->nsp + r->ndp) * sizeof(unsigned int));
 
 		qsort(lp, nlp, sizeof(unsigned int), int_order);
 		qsort(rp, nrp, sizeof(unsigned int), int_order);
@@ -1127,12 +1191,12 @@ void list_file(char *path, int acct)
 
 		rec->acct = acct;
 		sscanf(buf,
-		    "%lX/%lX->%lX/%lX %X %u %u %lu %lu %u %u %u %u %u %u %u %u %u %u",
-		    &rec->sa, &rec->sm, &rec->da, &rec->dm,
-		    &rec->flags, &rec->nsp, &rec->ndp, &rec->npkt, &rec->nbyt,
-		    &rec->ports[0], &rec->ports[1], &rec->ports[2], &rec->ports[3],
-		    &rec->ports[4], &rec->ports[5], &rec->ports[6], &rec->ports[7],
-		    &rec->ports[8], &rec->ports[9]);
+		    "%lX/%lX->%lX/%lX %lX %X %u %u %lu %lu %u %u %u %u %u %u %u %u %u %u",
+		    &rec->sa, &rec->sm, &rec->da, &rec->dm, &rec->iface,
+		    &rec->fw_flg, &rec->nsp, &rec->ndp, &rec->npkt, &rec->nbyt,
+		    &rec->fw_pts[0], &rec->fw_pts[1], &rec->fw_pts[2], &rec->fw_pts[3],
+		    &rec->fw_pts[4], &rec->fw_pts[5], &rec->fw_pts[6], &rec->fw_pts[7],
+		    &rec->fw_pts[8], &rec->fw_pts[9]);
 	}
 	fclose(f);
 
@@ -1147,17 +1211,17 @@ void list_file(char *path, int acct)
 
 	for (rec = recs; nused-- > 0; ++rec)
 	{
-		unsigned int *pp = &rec->ports[0];
+		unsigned int *pp = &rec->fw_pts[0];
 
 		if (!rec->acct)
 		{
 #ifdef DO_MASQUERADE		
-			if (rec->flags & IP_FW_F_MASQ)
+			if (rec->fw_flg & IP_FW_F_MASQ)
 				printf("(masquerade");
 			else
 #endif
 			{
-				if (rec->flags & IP_FW_F_ACCEPT)
+				if (rec->fw_flg & IP_FW_F_ACCEPT)
 					printf("accept\t");
 				else
 					printf("deny\t");
@@ -1181,7 +1245,7 @@ void list_file(char *path, int acct)
 				printf("%6lu \t", rec->nbyt);
 		}
 
-		switch (rec->flags & IP_FW_F_KIND)
+		switch (rec->fw_flg & IP_FW_F_KIND)
 		{
 		case IP_FW_F_ALL:
 			printf("all   ");
@@ -1200,16 +1264,16 @@ void list_file(char *path, int acct)
 		printf("%-19.19s ", addr_to_name(rec->sa, rec->sm));
 		printf("%-19.19s ", addr_to_name(rec->da, rec->dm));
 
-		if (rec->flags & (IP_FW_F_TCP | IP_FW_F_UDP))
+		if (rec->fw_flg & (IP_FW_F_TCP | IP_FW_F_UDP))
 		{
 			char *sep = "";
-			char *proto = (rec->flags & IP_FW_F_TCP) ? "tcp" : "udp";
+			char *proto = (rec->fw_flg & IP_FW_F_TCP) ? "tcp" : "udp";
 
 			if (rec->nsp == 0)
 				printf("any");
 			else
 			{
-				if (rec->flags & IP_FW_F_SRNG)
+				if (rec->fw_flg & IP_FW_F_SRNG)
 				{
 					printf("%u-%u", pp[0], pp[1]);
 					sep = ",";
@@ -1235,7 +1299,7 @@ void list_file(char *path, int acct)
 				printf("any");
 			else
 			{
-				if (rec->flags & IP_FW_F_DRNG)
+				if (rec->fw_flg & IP_FW_F_DRNG)
 				{
 					printf("%u-%u", pp[0], pp[1]);
 					sep = ",";
@@ -1300,6 +1364,17 @@ char **argv;
 		perror("ipfw: raw socket creation");
 		exit(1);
 	}
+
+	/* -d disables actions and debugs */
+	
+	if (argv[1] && strcmp(argv[1], "-d")==0)	
+	{
+		argv++;
+		argc--;
+		close(socket_fd);
+		socket_fd = -1;
+	}
+	
 	/* -n disables nameserver lookups */
 
 	if (argv[1] && strcmp(argv[1], "-n")==0)
@@ -1318,6 +1393,33 @@ char **argv;
 		type = "list";
 		list(socket_fd, &argv[2]);
 	}
+	else if(MATCH(argv[1], "policy"))
+	{
+		int type;
+		int mode;
+		if(MATCH(argv[2],"blocking"))
+			type=IP_FW_POLICY_BLK;
+		else if(MATCH(argv[2], "forwarding"))
+			type=IP_FW_POLICY_FWD;
+		else
+		{
+			fprintf(stderr,"ipfw: expected \"blocking\" or \"forwarding\".\n");
+			exit(1);
+		}
+		if(MATCH(argv[3], "reject"))
+			mode=IP_FW_F_ICMPRPL;
+		else if(MATCH(argv[3], "accept"))
+			mode=IP_FW_F_ACCEPT;
+		else if(MATCH(argv[3], "deny"))
+			mode=0;
+		else
+		{	
+			fprintf(stderr,"ipfw: expected \"accept\", \"deny\" or \"reject\".\n");
+			exit(1);
+		}
+		do_setsockopt(argv[2], socket_fd, IPPROTO_IP, type, &mode, sizeof(mode),0);
+		exit(0);
+	}
 	else if (MATCH(argv[1], "flush"))
 	{
 		if (MATCH(argv[2], "accounting"))
@@ -1331,12 +1433,18 @@ char **argv;
 		{
 			/* Same kludge as above, see above ranting and griping -BB */
 			unsigned long fred = 1;
-			(void) do_setsockopt(argv[1], socket_fd, IPPROTO_IP, IP_FW_FLUSH, &fred, sizeof(unsigned long), 0);
+			(void) do_setsockopt(argv[1], socket_fd, IPPROTO_IP, IP_FW_FLUSH_FWD, &fred, sizeof(unsigned long), 0);
+		}
+		else if (MATCH(argv[2], "blocking"))
+		{
+			/* Same kludge as above, see above ranting and griping -BB */
+			unsigned long fred = 1;
+			(void) do_setsockopt(argv[1], socket_fd, IPPROTO_IP, IP_FW_FLUSH_BLK, &fred, sizeof(unsigned long), 0);
 		}
 		else
 		{
 			fprintf(stderr,
-			    "ipfw: expected \"accounting\" or \"firewall\".");
+			    "ipfw: expected \"accounting\", \"blocking\" or \"firewall\".");
 			exit(1);
 
 
